from launch.schemas.configuration.base import Configuration
from launch.schemas.configuration.graph import GraphConfiguration
from launch.schemas.configuration.scheduling import SchedulingConfiguration


class LaunchConfiguration(Configuration):
    graph_config: GraphConfiguration
    scheduling_config: SchedulingConfiguration

    def __init__(self, graph_config: GraphConfiguration, scheduling_config: SchedulingConfiguration):
        self.graph_config = graph_config
        self.scheduling_config = scheduling_config

    def dump_config_params(self):
        return self.graph_config.dump_config_params(), self.scheduling_config.dump_config_params()



import os
from abc import ABC


class Configuration(ABC):
    __root_folder = 'ksg_scheduling'

    def __getattr__(self, item):
        if item in self.__dict__:
            return self.__dict__[item]
        return None

    def project_path_to_relative(self, path: str):
        return path
        """
        Converts a path relative to the project root to a path relative to the execution root
        :param path: Project root relative path
        :return: Execution root relative path
        """
        sep = os.sep if path.count(os.sep) >= path.count(os.altsep) else os.altsep
        execution_nest = os.getcwd().split(self.__root_folder)[1].count(os.sep)
        return f'..{sep}' * execution_nest + path.lstrip(sep)

    def dump_config_params(self) -> tuple:
        ...



from enum import auto, Enum
from random import Random
from typing import Union

from pandas import DataFrame
from sampo.generator.pipeline.project import SyntheticGraphType
from sampo.schemas.contractor import ContractorName
from sampo.schemas.types import WorkerName
from sampo.utilities.task_name import NameMapper
from sampo.utilities.visualization.base import VisualizationMode

from external.contractor_type import ContractorType
from external.estimate_time import WorkResourceEstimator
from external.lib.contractor_team import ContractorTeam
from launch.schemas.configuration.base import Configuration


class GraphConfigurationType(Enum):
    Generate = auto()
    ReadGraphGenerateResources = auto()
    ReadGraph = auto()


class GraphConfiguration(Configuration):
    type: GraphConfigurationType
    graph_info_file: Union[str, DataFrame] | None
    use_graph_lag_optimization: bool | None
    graph_fig_visualization: VisualizationMode | None
    generate_input: bool | None
    graph_mode: SyntheticGraphType | None
    synthetic_graph_vertices_lower_bound: int | None
    contractor_type: ContractorName | None
    contractor_team: WorkerName | None
    use_task_resource_generation: bool | None
    unique_work_names_mapper: NameMapper | None
    work_resource_estimator: WorkResourceEstimator | None

    def __init__(self,
                 graph_info_file: Union[str, DataFrame] | None = None,
                 use_graph_lag_optimization: bool | None = None,
                 graph_fig_visualization: VisualizationMode | None = VisualizationMode.NoFig,
                 generate_input: bool | None = None,
                 rand: Random | None = None,
                 graph_mode: SyntheticGraphType | None = None,
                 synthetic_graph_vertices_lower_bound: int | None = None,
                 contractor_type: ContractorName | None = None,
                 contractor_team: WorkerName | None = None,
                 use_task_resource_generation: bool | None = None,
                 unique_work_names_mapper: NameMapper | None = None,
                 work_resource_estimator: WorkResourceEstimator | None = None):
        self.graph_info_file = graph_info_file
        self.use_graph_lag_optimization = use_graph_lag_optimization
        self.graph_fig_visualization = graph_fig_visualization
        self.generate_input = generate_input
        self.rand = rand
        self.graph_mode = graph_mode
        self.synthetic_graph_vertices_lower_bound = synthetic_graph_vertices_lower_bound
        self.contractor_type = contractor_type
        self.contractor_team = contractor_team
        self.use_task_resource_generation = use_task_resource_generation
        self.unique_work_names_mapper = unique_work_names_mapper
        self.work_resource_estimator = work_resource_estimator

    def with_visualization(self, visualization_mode: VisualizationMode) -> 'GraphConfiguration':
        self.graph_fig_visualization = visualization_mode
        return self

    @staticmethod
    def generate_graph(graph_mode: SyntheticGraphType,
                       use_graph_lag_optimization: bool,
                       contractor_capacity: list[int] = [ContractorType.Average.command_capacity()],
                       contractor_team: list[str] = [ContractorTeam.Multi.resources_list()],
                       synthetic_graph_vertices_lower_bound: int | None = 200,
                       rand: Random | None = None) -> 'GraphConfiguration':
        config = GraphConfiguration(generate_input=True,
                                    graph_mode=graph_mode,
                                    use_graph_lag_optimization=use_graph_lag_optimization,
                                    contractor_type=contractor_capacity,
                                    contractor_team=contractor_team,
                                    synthetic_graph_vertices_lower_bound=synthetic_graph_vertices_lower_bound,
                                    rand=rand)
        config.type = GraphConfigurationType.Generate
        return config

    @staticmethod
    def read_graph_generate_resources(graph_info: Union[str, DataFrame],
                                      use_graph_lag_optimization: bool,
                                      work_resource_estimator: WorkResourceEstimator,
                                      contractor_capacity: list[int] = [ContractorType.Average.command_capacity()],
                                      contractor_resources: list[str] = [ContractorTeam.Multi.resources_list()],
                                      unique_work_names_mapper: NameMapper | None = None) -> 'GraphConfiguration':
        config = GraphConfiguration(use_task_resource_generation=True,
                                    graph_info_file=graph_info,
                                    use_graph_lag_optimization=use_graph_lag_optimization,
                                    contractor_type=contractor_capacity,
                                    contractor_team=contractor_resources,
                                    work_resource_estimator=work_resource_estimator,
                                    unique_work_names_mapper=unique_work_names_mapper)
        config.type = GraphConfigurationType.ReadGraphGenerateResources
        return config

    @staticmethod
    def read_graph(graph_info_filepath: str,
                   use_graph_lag_optimization: bool,
                   work_resource_estimator: WorkResourceEstimator | None = None) -> 'GraphConfiguration':
        config = GraphConfiguration(graph_info_file=graph_info_filepath,
                                    use_graph_lag_optimization=use_graph_lag_optimization,
                                    work_resource_estimator=work_resource_estimator)
        config.type = GraphConfigurationType.ReadGraph
        return config

    def dump_config_params(self) -> tuple:
        return (self.project_path_to_relative(self.graph_info_file)
                    if isinstance(self.graph_info_file, str)
                    else self.graph_info_file,
                self.use_graph_lag_optimization,
                self.graph_fig_visualization,
                self.generate_input,
                self.rand,
                self.graph_mode,
                self.synthetic_graph_vertices_lower_bound,
                self.contractor_type,
                self.contractor_team,
                self.use_task_resource_generation,
                self.unique_work_names_mapper,
                self.work_resource_estimator)



from dataclasses import dataclass

from sampo.metrics.resources_in_time.service import ResourceOptimizationType
from sampo.scheduler.base import SchedulerType
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph
from sampo.schemas.serializable import AutoJSONSerializable
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.utilities.serializers import custom_serializer
from sampo.utilities.task_name import NameMapper
from sampo.utilities.visualization.base import VisualizationMode
from sampo.utilities.visualization.resources import EmploymentFigType

from external.estimate_time import WorkTimeEstimationMode
from launch.schemas.configuration.base import Configuration


@dataclass
class SchedulingConfiguration(Configuration):
    input_configuration: 'SchedulingInputConfiguration'
    inner_configuration: 'SchedulingInnerConfiguration'
    output_configuration: 'SchedulingOutputConfiguration'

    def __getattr__(self, item):
        for member in self.input_configuration, self.output_configuration, self.inner_configuration:
            it = member.__getattr__(item)
            if it is not None:
                break
        return it

    def dump_config_params(self) -> tuple:
        return self.input_configuration.dump_config_params() \
            + self.inner_configuration.dump_config_params() \
            + self.output_configuration.dump_config_params()


@dataclass
class SchedulingInputConfiguration(AutoJSONSerializable['SchedulingInputConfiguration'], Configuration):
    work_graph: WorkGraph
    contractors: list[Contractor]

    @custom_serializer('contractors')
    def contractors_serializer(self, value):
        return [c._serialize() for c in value]

    @classmethod
    @custom_serializer('contractors', deserializer=True)
    def contractors_deserializer(cls, value):
        return [Contractor._deserialize(c) for c in value]

    def dump_config_params(self) -> tuple:
        return self.work_graph, self.contractors


class SchedulingInnerConfiguration(Configuration):
    start: str
    end: str
    algorithm_type: SchedulerType
    resource_optimization: ResourceOptimizationType | None
    use_idle_estimator: bool
    validate_schedule: bool
    deadline: Time | None
    estimator_mode: WorkTimeEstimationMode | None
    inverse_name_mapper: NameMapper | None
    work_time_estimator: WorkTimeEstimator | None

    def __init__(self, algorithm_type: SchedulerType, start: str, end: str, validate_schedule: bool):
        self.algorithm_type = algorithm_type
        self.start = start
        self.end = end
        self.validate_schedule = validate_schedule

        self.deadline = None
        self.estimator_mode = None
        self.resource_optimization = None
        self.inverse_name_mapper = None
        self.work_time_estimator = None

    def with_work_time_estimator(self, work_time_estimator: WorkTimeEstimator,
                                 use_idle_estimator: bool,
                                 estimator_mode: WorkTimeEstimationMode | None = WorkTimeEstimationMode.Realistic) \
            -> 'SchedulingInnerConfiguration':
        self.work_time_estimator = work_time_estimator
        self.use_idle_estimator = use_idle_estimator
        self.estimator_mode = estimator_mode
        return self

    def with_deadline_resource_optimization(self, deadline: Time, optimization_type: ResourceOptimizationType) \
            -> 'SchedulingInnerConfiguration':
        self.deadline = deadline
        self.resource_optimization = optimization_type
        return self

    def with_inverse_name_mapper(self, inverse_name_mapper: NameMapper) -> 'SchedulingInnerConfiguration':
        self.inverse_name_mapper = inverse_name_mapper
        return self

    def dump_config_params(self) -> tuple:
        return (self.start,
                self.end,
                self.deadline,
                self.algorithm_type,
                self.resource_optimization,
                self.work_time_estimator,
                self.use_idle_estimator,
                self.estimator_mode,
                self.inverse_name_mapper,
                self.validate_schedule)


class SchedulingOutputConfiguration(Configuration):
    title: str
    output_folder_path: str
    save_to_xer: bool
    save_to_csv: bool
    save_to_json: bool
    gant_chart_visualization: VisualizationMode
    employment_figs: list[EmploymentFigType | VisualizationMode] | None

    def __init__(self, title: str, output_folder: str, save_to_xer: bool, save_to_csv: bool, save_to_json: bool):
        self.title = title
        self.output_folder_path = output_folder
        self.save_to_xer = save_to_xer
        self.save_to_csv = save_to_csv
        self.save_to_json = save_to_json

        self.employment_figs = None

    def with_gant_chart(self, visualization_mode: VisualizationMode) -> 'SchedulingOutputConfiguration':
        self.gant_chart_visualization = visualization_mode
        return self

    def with_one_more_employment_fig(self, fig_type: EmploymentFigType, visualization_mode: VisualizationMode) \
            -> 'SchedulingOutputConfiguration':
        if self.employment_figs is None:
            self.employment_figs = list()
        self.employment_figs.append((fig_type, visualization_mode))
        return self

    def dump_config_params(self) -> tuple:
        return (self.title,
                self.project_path_to_relative(self.output_folder_path),
                self.save_to_xer,
                self.save_to_csv,
                self.save_to_json,
                self.gant_chart_visualization,
                self.employment_figs)



import pandas as pd
import ast
from typing import Dict

from sampo.schemas.time import Time


def fix_split_tasks(baps_schedule_df: pd.DataFrame) -> pd.DataFrame:
    """
    Process and merge information for all tasks, which were separated on the several stages during split
    :param baps_schedule_df: pd.DataFrame: schedule with info for tasks separated on stages
    :return: pd.DataFrame: schedule with merged info for all real tasks
    """
    df_len = baps_schedule_df.shape[0]
    baps_schedule_df.index = range(df_len)

    df = pd.DataFrame(columns=baps_schedule_df.columns)
    unique_ids = set([x.split('_')[0] for x in baps_schedule_df.task_id])

    for task_id in unique_ids:
        task_stages_df = baps_schedule_df.loc[baps_schedule_df.loc[:, 'task_id'].str.contains(task_id)]
        task_series = merge_split_stages(task_stages_df.reset_index(drop=True))
        df.loc[df.shape[0]] = task_series  # append

    df = df.sort_values(by=['start', 'task_name'])
    df['idx'] = range(len(df))

    return df


def merge_split_stages(task_df: pd.DataFrame) -> pd.Series:
    """
    Merge split stages of the same real task into one
    :param task_df: pd.DataFrame: one real task's stages dataframe, sorted by start time
    :return: pd.Series with the full information about the task
    """
    if len(task_df) == 1:
        df = task_df.copy()
        df['successors'] = [[tuple([x[0].split('_')[0], x[1]]) for x in df.loc[0, 'successors']]]
        return df.loc[0, :]
    else:
        df = task_df.copy()
        df = df.iloc[-1:].reset_index(drop=True)
        for column in ['task_id', 'task_name']:
            df.loc[0, column] = df.loc[0, column].split('_')[0]  # fix task id and name

        # sum up volumes through all stages
        df.loc[0, 'volume'] = sum(task_df.loc[:, 'volume'])
        df.loc[0, 'workers'] = task_df.loc[0, 'workers']

        # fix connections through all stages
        fixed_connections_lst = []
        for connections_lst in task_df.loc[:, 'successors']:
            for connection in connections_lst:
                if connection[1] != 'IFS':
                    fixed_connections_lst.append(tuple([connection[0].split('_')[0], connection[1]]))
        fixed_connections_lst = list(set(fixed_connections_lst))
        df.loc[:, 'successors'] = [fixed_connections_lst]

        # fix task's start time and duration
        df.loc[0, 'start'] = task_df.loc[0, 'start']
        df.loc[0, 'finish'] = task_df.loc[len(task_df) - 1, 'finish']
        df.loc[0, 'duration'] = (df.loc[0, 'finish'] - df.loc[0, 'start']).days + 1

        return df.loc[0, :]


def remove_service_tasks(service_schedule_df: pd.DataFrame) -> pd.DataFrame:
    """
    Remove 'start', 'finish' and milestone tasks from the schedule
    :param service_schedule_df: pd.DataFrame: schedule (with merges stages in the case of baps) with service tasks
    :return: pd.DataFrame: schedule without information about service tasks
    """
    schedule_df = service_schedule_df.copy()

    service_df = schedule_df.loc[:, 'task_name'].str.contains('start|finish')

    # Prepare list with service tasks ids
    service_tasks_ids = set(schedule_df.loc[service_df].loc[:, 'task_id'])

    # Remove rows with service tasks from DataFrame
    schedule_df = schedule_df.loc[~service_df]

    # Fix connections linked to the service tasks
    fixed_connections_lst = []
    for connections_lst in schedule_df.loc[:, 'successors']:
        fixed_connections_lst.append([])
        for connection in connections_lst:
            if connection[0] not in service_tasks_ids:
                fixed_connections_lst[-1].append(connection)
    schedule_df.loc[:, 'successors'] = fixed_connections_lst
    return schedule_df


def get_full_resources_dict(workers_sets: list[Dict[str, float]]) -> Dict[str, int]:
    """
    Prepare resources dict
    :param workers_sets: list with info about assigned resources for tasks
    :return: dict with resources names as keys and resources ids as values
    """
    resources_names = []
    for workers_set in workers_sets:
        resources_names.extend(list(workers_set.keys()))
    resources_names = list(set(resources_names))
    resources_ids = range(len(resources_names))
    return dict(zip(resources_names, resources_ids))


# Main function for converting schedules to validation jsons
def schedule_to_json(schedule_uploading_mode='from_df', schedule_file_path='', schedule_df=None,
                     schedule_name='schedule', deadline_date='2030-01-01') -> Dict:
    """
    Create dictionary with the JSON structure by given schedule and additional ksg info
    :param schedule_uploading_mode: one of two: 'from_df' (schedule_df needed)
    or 'from_file' (schedule_file_path needed)
    :param schedule_file_path: str path to the schedule .csv file
    :param schedule_df: pandas.Dataframe
    :param schedule_name: str
    :return: dict with JSON structure of tasks, assigned execution times and resources
    """
    if schedule_uploading_mode == 'from_file':
        df = pd.read_csv(schedule_file_path, sep=';')
    elif schedule_uploading_mode == 'from_df':
        df = schedule_df.copy()
        if not isinstance(schedule_df, pd.DataFrame):
            raise Exception("schedule_df attribute should have type pandas.DataFrame, got " +
                            str(type(schedule_df)) + " instead")
    else:
        raise Exception("unknown schedule uploading mode")

    df.loc[:, 'workers'] = [ast.literal_eval(x) for x in df.loc[:, 'workers']]
    resources_dict = get_full_resources_dict(df.loc[:, 'workers'])

    schedule_dict = {"plan_name": schedule_name, "plan_deadline": deadline_date, "activities": []}
    for i in df.index:
        if not ('start' in df.loc[i, 'task_name'] or 'finish' in df.loc[i, 'task_name']):
            activity_dict = {"activity_id": str(df.loc[i, 'task_id']),
                             "activity_name": df.loc[i, 'task_name'],
                             "start_date": str(df.loc[i, 'start']), "end_date": str(df.loc[i, 'finish']),
                             "volume": str(df.loc[i, 'volume']),
                             "measurement": str(df.loc[i, 'measurement']),
                             "labor_resources": [], "non_labor_resources": {},
                             "descendant_activities": df.loc[i, 'successors']}
            for worker in df.workers[i]:
                labor_info = {"labor_id": int(resources_dict[worker]), "labor_name": worker,
                              "volume": float(df.loc[i, 'workers'][worker]), "workers": []}
                worker_info = {"worker_id": 0, "contractor_id": 0, "contractor_name": 'Main contractor',
                               "working_periods": [{"start_datetime": str(df.loc[i, 'start']),
                                                    "end_datetime": str(df.loc[i, 'finish']),
                                                    "productivity": 1}]}
                labor_info["workers"].append(worker_info)
                activity_dict["labor_resources"].append(labor_info)
            schedule_dict["activities"].append(activity_dict)
    return schedule_dict